#ifndef CADICAL_SOLVER_INTERFACE_H
#define CADICAL_SOLVER_INTERFACE_H

#include "global.h"
#include "solveGeneral.hpp"
#include "cadical.hpp"
#include <bits/stdc++.h>
#include <algorithm>

class CadicalSolver : public CommonInterface, public CaDiCaL::ExternalPropagator
{
private:
    CaDiCaL::Solver *solver;
    const std::vector<int> *model;
    bool redundant;

    bool changeInTrail = true; // checks whether the trail has changed since the last propagation step

    int highestEdgeVariable;
    vector<pair<int, int>> lit2edge; // get for positive literal the corresponding edge

    vector<vector<int>> clauses;
    int highestVariable;
    int checkMode = false; // if true solver has finished and clauses are added by the normal "incremental interface"
    adjacency_matrix_t currentMatrix;
    vector<vector<bool>> fixedMatrix; // true if entry in currentMatrix is fixed.
    deque<vector<int>> current_trail; // for each decision lvl store the assigned literals (only positive version)

    vector<vector<int>> literal2clausePos; // for each edge variable store clause which was used the last time.
    vector<vector<int>> literal2clauseNeg; // for each negation of an edge variable
public:
    CadicalSolver(cnf_t &cnf, int highestVariable);
    ~CadicalSolver() { solver->disconnect_external_propagator(); }

protected: // virtual classes from common interface
    void solve(vector<int> assumptions);
    bool solve(vector<int> assumptions, int timeout);

    adjacency_matrix_t getAdjacencyMatrix()
    {
        // printf("Trail: ");
        // for (auto lit: *current_trail)
        //     printf("%d ", lit);
        // printf("\n");
        if (!checkMode)
        {
            return currentMatrix;
        }
        else
        {
            adjacency_matrix_t matrix(vertices, vector<adjacencyState_t>(vertices, unknown));
#ifndef DIRECTED
            for (int i = 0; i < vertices; i++)
                for (int j = i + 1; j < vertices; j++)
                    matrix[i][j] = matrix[j][i] = solver->val(edges[i][j]) > 0 ? adjacent : not_adjacent;
#else
            for (int i = 0; i < vertices; i++)
                for (int j = 0; j < vertices; j++)
                    if (i != j)
                        matrix[i][j] = solver->val(edges[i][j]) > 0 ? adjacent : not_adjacent;

#endif
            return matrix;
        }
    }

    void addClause(const vector<lit_t> &clause, bool redundant)
    {
        if (addedClauses)
        {
            for (auto l : clause)
                fprintf(addedClauses, "%d ", l);
            fprintf(addedClauses, "0\n");
        }
        // if (sym_breaking_clause.size() != 0)
        //    return; // EXIT_UNWANTED_STATE
        // printf("Number of literals: %ld, Add the following clause:", clause.size());
        // for (auto lit : clause)
        //     printf("%d ", lit);
        // printf("\n");
        if (!checkMode)
            this->clauses.push_back(clause);
        else
        {
            // use incremental interface
            for (auto l : clause)
                solver->add(l);
            solver->add(0);
        }
    }

public:
    void notify_assignment(int lit, bool is_fixed)
    {
        changeInTrail = true;
        int absLit = abs(lit);
        if (!is_fixed) // push back literal to undo if current decission literal is changed
            current_trail.back().push_back(absLit);
        auto edge = lit2edge[absLit];
        if (lit > 0)
        {
            currentMatrix[edge.first][edge.second] = adjacent;
        }
        else
        {
            currentMatrix[edge.first][edge.second] = not_adjacent;
        }
        if (is_fixed)
            fixedMatrix[edge.first][edge.second] = true;
#ifndef DIRECTED
        currentMatrix[edge.second][edge.first] = currentMatrix[edge.first][edge.second];
#endif
    }

    void notify_new_decision_level()
    {
        current_trail.push_back(vector<int>());
    }

    void notify_backtrack(size_t new_level)
    {
        while (current_trail.size() > new_level + 1)
        {
            auto last = current_trail.back();
            for (int l : last)
            {
                pair<int, int> edge = lit2edge[l];
                if (fixedMatrix[edge.first][edge.second])
                    continue;
                currentMatrix[edge.first][edge.second] = unknown;
#ifndef DIRECTED
                currentMatrix[edge.second][edge.first] = unknown;
#endif
            }
            current_trail.pop_back();
        }
    }

    // currently not checked in propagator but with the normal incremental interface to allow adding other literals or even new once.
    bool cb_check_found_model(const std::vector<int> &model)
    {
        if (!clauses.empty())
            return false; // EXIT_UNWANTED_STATE only do check if there isn't another clause to add before
        // this->current_trail = &model;
        if (chechSolutionInProp)
        {
            return check();
        }
        return true;
    }

    bool check_solution()
    {
        if (!chechSolutionInProp)
        {
            checkMode = true;
            bool res = check();
            checkMode = false;
            return res;
        }
        return true;
    }

    bool cb_has_external_clause()
    {
        // PRINT_CURRENT_LINE
        // if no clause, then check whether a clause could be added. If already a clause present then just return clause.
        // if propagation is done in other function then not compute clauses here
        if (clauses.empty() && changeInTrail && !propagateLiteralsCadical)
        {
            changeInTrail = false;
            propagate();
        }

        // printf("Check for external clause: %ld\n", sym_breaking_clause.size());
        return !clauses.empty();
    }

    int cb_add_external_clause_lit()
    {
        // PRINT_CURRENT_LINE
        // printf("Call: Add external clause\n");
        vector<int> &lastClause = clauses[clauses.size() - 1];
        if (lastClause.empty())
        {
            clauses.pop_back(); // delete last clause
            // printf(" end clause\n");
            return 0;
        }
        else
        {
            // printf("Add external clause\n");
            int lit = lastClause.back();
            lastClause.pop_back();
            // printf("%d ", lit);
            return lit;
        }
    }

    // functions need to be defined
    int cb_decide() { return 0; }

    int cb_propagate()
    {
        // PRINT_CURRENT_LINE
        if (!propagateLiteralsCadical)
            return 0;

        // PRINT_CURRENT_LINE
        
        if (!changeInTrail)
            return 0;

        // PRINT_CURRENT_LINE

        // TODO changeInTrail = false;
        propagate();

        if (clauses.empty())
            return 0;

        auto lastClause = clauses.back();
        assert(!lastClause.empty());
        // find unassigned literal otherwise take last one; first check if clause is unit
        int nUnknown = 0;
        int unassigned = 0;
        for (auto l : lastClause)
        {
            auto edge = lit2edge[abs(l)];
            if (currentMatrix[edge.first][edge.second] == unknown)
            {
                nUnknown++;
                unassigned = l;
            }
            else if (currentMatrix[edge.first][edge.second] == adjacent && l > 0)
                return 0; // already satisfied
            else if (currentMatrix[edge.first][edge.second] == not_adjacent && l < 0)
                return 0; // already satisfied
        }

        if (nUnknown == 1)
        {
            // notify_assignment(unassigned, false); // push back the clause to the current trail
            clauses.pop_back(); // delete last clause
            if (unassigned > 0)
                literal2clausePos[abs(unassigned)] = lastClause;
            else
                literal2clauseNeg[abs(unassigned)] = lastClause;
            // PRINT_CURRENT_LINE
            return unassigned;
        }
        return 0;
    }

    int cb_add_reason_clause_lit(int plit)
    {
        // PRINT_CURRENT_LINE
        if (plit > 0)
        {
            if (literal2clausePos[abs(plit)].empty())
                return 0;
            auto l = literal2clausePos[abs(plit)].back();
            literal2clausePos[abs(plit)].pop_back();
            return l;
        }
        else
        {
            if (literal2clauseNeg[abs(plit)].empty())
                return 0;
            auto l = literal2clauseNeg[abs(plit)].back();
            literal2clauseNeg[abs(plit)].pop_back();
            return l;
        }
    };
};

#endif
