/**
 * Implementation of common parts amongst all solvers
 */
#include <boost/graph/boyer_myrvold_planar_test.hpp>
#include <boost/graph/is_kuratowski_subgraph.hpp>
#include <boost/graph/adjacency_list.hpp>
#include "useful.h"
#include "minimalityCheck.h"
#include "global.h"
#include "solve.h"
#include "path.h"
#include "connectedComponents.h"
#include "coloring.h"
#include "filter.hpp"
#include "solveGeneral.hpp"
#include "cadical.hpp"

typedef int lit_t;

void initEdgeMemory(int nVertices)
{
  edge_stats.resize(nVertices);
  for (int v = 0; v < nVertices; v++)
  {
    edge_stats[v].resize(nVertices, 0);
  }
}

void initTriangleMemory(int nVertices)
{
  triangle_stats.resize(nVertices);
  for (int v = 0; v < nVertices; v++)
  {
    triangle_stats[v].resize(nVertices);
    for (int w = 0; w < nVertices; w++)
    {
      triangle_stats[v][w].resize(nVertices, 0);
    }
  }
}

void printEdgeStats()
{
  printf("Edge occurrence statistics:\n");
  for (int u = 0; u < vertices; u++)
  {
    for (int v = 0; v < vertices; v++)
    {
      printf("%4u ", edge_stats[u][v]);
    }
    printf("\n");
  }
}

vector<lit_t> monochromaticEdges2Clause(const vector<pair<int, int>> &mches)
{
  vector<lit_t> clause;
  if (hyperedgeColoring)
  {
    for (auto e : mches)
    {
      clause.push_back(edges_intersection_graph[e.first][e.second]);
    }
  }
  else
  {
    for (auto e : mches)
    {
      clause.push_back(edges[e.first][e.second]);
    }
  }
  return clause;
}

// computes the adjacency matrix of the intersection graph
// of the hypergraph given by the incidence matrix
adjacency_matrix_t getIntersectionMatrix(const adjacency_matrix_t &incidence_matrix)
{
  adjacency_matrix_t matrix(b_vertices[1], vector<adjacencyState_t>(b_vertices[1]));

  for (int e = 0; e < b_vertices[1]; e++)
  {
    for (int f = e + 1; f < b_vertices[1]; f++)
    {
      matrix[e][f] = matrix[f][e] = not_adjacent;
      for (int v = 0; v < b_vertices[0]; v++)
      {
        if (incidence_matrix[v][e + b_vertices[0]] && incidence_matrix[v][f + b_vertices[0]])
        {
          matrix[e][f] = matrix[f][e] = adjacent;
          break;
        }
      }
    }
  }
  return matrix;
}

// given the incidence matrix, the function return a clique of the intersection matrix
vector<int> getCliqueInIntersectionMatrix(const adjacency_matrix_t &incidence_matrix)
{
  adjacency_matrix_t matrix(b_vertices[1], vector<adjacencyState_t>(b_vertices[1]));
  // compute vertex contained in the most hyperedges
  int maxVertex = 0;
  int maxEdges = 0;
  for (int v = 0; v < b_vertices[0]; v++)
  {
    int counter = 0;
    for (int e = 0; e < b_vertices[1]; e++)
      if (incidence_matrix[v][e + b_vertices[0]] == adjacent)
        counter++;
    if (counter > maxEdges)
    {
      maxEdges = counter;
      maxVertex = v;
    }
  }
  vector<int> clique;
  for (int e = 0; e < b_vertices[1]; e++)
    if (incidence_matrix[maxVertex][e + b_vertices[0]] == adjacent)
      clique.push_back(e);

  int clique_idx = 0;
  int clique_size = clique.size();
  for (int e = 0; e < b_vertices[1]; e++)
  {
    while (clique_idx < clique_size && clique[clique_idx] < e)
      clique_idx++;
    if (clique_idx < clique_size && e == clique[clique_idx])
    {
      clique_idx++;
      continue;
    }
    bool can_be_added = true;
    for (int f : clique)
    {
      bool intersects = false;
      for (int v = 0; v < b_vertices[0]; v++)
      {
        if (incidence_matrix[v][e + b_vertices[0]] == adjacent && incidence_matrix[v][f + b_vertices[0]] == adjacent)
        {
          intersects = true;
          break;
        }
      }
      if (!intersects)
      {
        can_be_added = false;
        break;
      }
    }
    if (can_be_added)
    {
      clique.push_back(e);
    }
  }

  return clique;
}

// returns either an edge list containing an kuratowski subgraph or an empty list if planar; checks upper triangular matrix
vector<pair<int, int>> testPlanarity(adjacency_matrix_t &m)
{
  int n = vertices; // vertices is already used in the boost namespace
  using namespace boost;
  typedef adjacency_list<vecS,
                         vecS,
                         undirectedS,
                         property<vertex_index_t, int>,
                         property<edge_index_t, int>>
      graph;

  // Create a K_6 (complete graph on 6 vertices), which
  // contains both Kuratowski subgraphs as minors.
  graph g(n);
  for (int i = 0; i < n; i++)
    for (int j = i + 1; j < n; j++)
      if (m[i][j] == adjacent)
        add_edge(i, j, g);

  // Initialize the interior edge index
  property_map<graph, edge_index_t>::type e_index = get(edge_index, g);
  graph_traits<graph>::edges_size_type edge_count = 0;
  graph_traits<graph>::edge_iterator ei, ei_end;
  for (boost::tie(ei, ei_end) = boost::edges(g); ei != ei_end; ++ei)
    put(e_index, *ei, edge_count++);

  // Test for planarity - we know it is not planar, we just want to
  // compute the kuratowski subgraph as a side-effect
  typedef std::vector<graph_traits<graph>::edge_descriptor>
      kuratowski_edges_t;
  kuratowski_edges_t kuratowski_edges;
  if (boyer_myrvold_planarity_test(boyer_myrvold_params::graph = g,
                                   boyer_myrvold_params::kuratowski_subgraph =
                                       std::back_inserter(kuratowski_edges)))
  {
    // std::cout << "Input graph is planar" << std::endl;
    return vector<pair<int, int>>();
  }
  else
  {
    // std::cout << "Input graph is not planar" << std::endl;
    // std::cout << "Edges in the Kuratowski subgraph: ";
    vector<pair<int, int>> kurEdges;
    kuratowski_edges_t::iterator ki, ki_end;
    ki_end = kuratowski_edges.end();
    for (ki = kuratowski_edges.begin(); ki != ki_end; ++ki)
    {
      // std::cout << *ki << " ";
      // printf(" %d %d", source(*ki, g), target(*ki, g));
      kurEdges.push_back(make_pair(source(*ki, g), target(*ki, g)));
    }
    // std::cout << std::endl;
    return kurEdges;

    // std::cout << "Is a kuratowski subgraph? ";
    // if (is_kuratowski_subgraph(g, kuratowski_edges.begin(), kuratowski_edges.end()))
    //   std::cout << "Yes." << std::endl;
    // else
    //   std::cout << "No." << std::endl;
  }
}

/**
 * @brief Check if graph represented by digraph has thickness 2.
 * The decomposition is represented by double edges for the first graph,
 * and single edges for the second one.
 *
 * @return true
 * @return false
 */
bool CommonInterface::thicknessTesting()
{
  auto m = getAdjacencyMatrix();
  auto m1 = vector<vector<adjacencyState_t>>(vertices, vector<adjacencyState_t>(vertices, not_adjacent));
  for (int i = 0; i < vertices; i++)
    for (int j = i + 1; j < vertices; j++)
      if (m[i][j] == adjacent && m[j][i] == adjacent)
        m1[i][j] = m1[j][i] = adjacent;

  auto edgesKur = testPlanarity(m1);
  if (edgesKur.size() != 0)
  {
    vector<int> clause;
    for (auto e : edgesKur)
    {
      clause.push_back(-edges[e.first][e.second]);
      clause.push_back(-edges[e.second][e.first]);
    }
    addClause(clause, true);
    return false;
  }

  // test lower triangular matrix
  auto m2 = vector<vector<adjacencyState_t>>(vertices, vector<adjacencyState_t>(vertices, not_adjacent));
  for (int i = 0; i < vertices; i++)
    for (int j = i + 1; j < vertices; j++)
      if (m[j][i] == adjacent && m[i][j] == not_adjacent) // only adjacent on lower side
        m2[i][j] = m2[j][i] = adjacent;
  edgesKur = testPlanarity(m2);
  if (edgesKur.size() != 0)
  {
    vector<int> clause;
    for (auto e : edgesKur)
    {
      clause.push_back(edges[min(e.first, e.second)][max(e.first, e.second)]);
      clause.push_back(-edges[max(e.first, e.second)][min(e.first, e.second)]);
    }
    addClause(clause, true);
    return false;
  }
  return true;
}

bool CommonInterface::planarityTesting()
{
  auto m = getAdjacencyMatrix();

  testPlanarity(m);
#ifndef DIRECTED
  auto edgesKur = testPlanarity(m);
#else
  // test undirected version for planarity
  auto m_undirected = m;
  for (int i = 0; i < vertices; i++)
    for (int j = i + 1; j < vertices; j++)
    {
      if (m[i][j] == adjacent || m[j][i] == adjacent)
        m_undirected[j][i] = m_undirected[i][j] = adjacent;
    }
  auto edgesKur = testPlanarity(m_undirected);
#endif

  if (edgesKur.size() != 0)
  {
    vector<int> clause;
    for (auto e : edgesKur)
    {
#ifndef DIRECTED
      clause.push_back(-edges[e.first][e.second]);
#else
      // clause.push_back(-edges[e.first][e.second]);
      if (m[e.first][e.second] == adjacent)
      {
        clause.push_back(-edges[e.first][e.second]);
      }
      else
      {
        clause.push_back(-edges[e.second][e.first]);
      }
#endif
    }
    addClause(clause, true);
    return false;
  }

  return true;
}

bool CommonInterface::checkMin(bool checkConnectedComponents)
{
  clock_t start = clock();
  bool res = true;
  adjacency_matrix_t matrix = getAdjacencyMatrix();

  // printAdjacencyMatrix(matrix);
  try
  {
    if (initialVertexOrderings.size() != 0)
    {
      for (auto ordering : initialVertexOrderings)
      {
        checkMinimality(matrix, ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
      }
    }
    else
    {
      vertex_ordering_t vertex_ordering(vertices);
      for (int i = 0; i < vertices; i++)
        vertex_ordering[i] = i;
      checkMinimality(matrix, vertex_ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
#ifndef DIRECTED
      if (checkComplement)
        checkMinimalityComplement(matrix, vertex_ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
#endif
    }

    if (!embeddedStarIntervalls.empty())
    {
      for (int i = 0; i < (int)embeddedStarIntervalls.size(); i++)
        for (int j = i + 1; j < (int)embeddedStarIntervalls.size(); j++)
        {
          // check swapping this two blocks
          vertex_ordering_t vertex_ordering(vertices);
          for (int i = 0; i < vertices; i++)
          {
            vertex_ordering[i] = i;
          }

          swap(vertex_ordering[embeddedStarIntervalls[i].first], vertex_ordering[embeddedStarIntervalls[j].first]);

          int minDegree = embeddedStar.first;
          // swap second neighbors
          for (int k = 0; k < minDegree - 1; k++)
          {
            swap(vertex_ordering[embeddedStarIntervalls[i].second.first + k], vertex_ordering[embeddedStarIntervalls[j].second.first + k]);
          }

          // for (auto v : vertex_ordering)
          //   printf("%d ", v);
          // printf("\n");

          checkMinimality(matrix, vertex_ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
        }
    }

    if (checkConnectedComponents) // special check
    {
      // check swapping paritions with same size
      int N = intervallsColoring.size(); // last partition is only the single vertex
      for (int i = 0; i < N; i++)
        for (int j = i + 1; j < N; j++)
          if (intervallsColoring[i].second - intervallsColoring[i].first == intervallsColoring[j].second - intervallsColoring[j].first)
          {
            // check swapping this two blocks
            vertex_ordering_t vertex_ordering(vertices);
            for (int i = 0; i < vertices; i++)
            {
              vertex_ordering[i] = i;
            }

            for (int k = 0; k < intervallsColoring[i].second - intervallsColoring[i].first; k++)
            {
              swap(vertex_ordering[k + intervallsColoring[i].first], vertex_ordering[k + intervallsColoring[j].first]);
            }
            checkMinimality(matrix, vertex_ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
          }

      // check swapping connected components
      checkSwaps(matrix);
    }

    if (fiveCycleVersion)
    {
      for (int i = 0; i < 5; i++) // starting vertex
      {
        vertex_ordering_t vertex_ordering(vertices);
        for (int i = 0; i < vertices; i++)
          vertex_ordering[i] = i;
        for (int j = 0; j < 5; j++)
          vertex_ordering[j] = (j + i) % 5;
        checkMinimality(matrix, vertex_ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
      }

      // reverse cycle
      for (int i = 0; i < 5; i++) // starting vertex
      {
        vertex_ordering_t vertex_ordering(vertices);
        for (int i = 0; i < vertices; i++)
          vertex_ordering[i] = i;
        for (int j = 0; j < 5; j++)
          vertex_ordering[5 - j - 1] = (j + i) % 5;
        checkMinimality(matrix, vertex_ordering, {.initial_partition = initialPartition, .cutoff = cutoff});
      }
    }
  }
  catch (LimitReachedException e)
  {
    printf("Limit reached\n");
  }
  catch (std::vector<signed_edge_t> e)
  {
    stats.nSymBreakClauses++;
    vector<lit_t> clause;
    for (auto signed_edge : e)
    {
      edge_t edge = signed_edge.second;
      if (signed_edge.first == adjacent)
        clause.push_back(edges[edge.first][edge.second]);
      else if (signed_edge.first == not_adjacent)
        clause.push_back(-edges[edge.first][edge.second]);

      // if (signed_edge.first == not_adjacent)
      //     printf("-");
      //   printf("(%d,%d) ", edge.first, edge.second);

      if (symBreakClausesFile)
      {
        if (signed_edge.first == not_adjacent)
          fprintf(symBreakClausesFile, "-");
        fprintf(symBreakClausesFile, "(%d,%d) ", edge.first, edge.second);
      }
    }
    // printf("\n");
    if (symBreakClausesFile)
    {
      fprintf(symBreakClausesFile, "\n");
    }

    addClause(clause, !irredundantSymClauses);
    res = false;
    // printf("Not minimal\n");
  }
  // printf("Time %f\n", ((double) clock() - start) / CLOCKS_PER_SEC);
  stats.timeMinimalityCheck += clock() - start;
  return res;
}

// returns true if cutoff was reached
bool CommonInterface::cutoffFunction()
{
  if (assignmentCutoffPrerun && assignmentCutoffPrerun > stats.callsPropagator)
    return false;

  if (assignmentCutoffPrerunTime && assignmentCutoffPrerunTime > (clock() - stats.start) / CLOCKS_PER_SEC)
    return false;

  assignmentCutoffPrerunTime = 0;
  assignmentCutoffPrerun = 0;
  frequency = 1; // check each time when creating cubes to add as much literals as possible

  adjacency_matrix_t matrix = getAdjacencyMatrix();

  if (assignmentScoring == COUNT_ASSIGNED)
  {
    int nAssigned = 0;
    for (int i = 0; i < vertices; i++)
      for (int j = i + 1; j < vertices; j++)
        if (matrix[i][j] != unknown)
          nAssigned++;

    if (nAssigned <= assignmentCutoff)
      return false;
  }
  else
  {
    // score = sum log_2 ( 1 / p_e )
    // where p_e is the empirically recorded frequency of e
    // for balanced variables that are true half the time,
    // the contribution towards the score is 1
    // this is supposed to measure the search-space reduction
    // in the number of bits
    double assignmentScore = 0.0;
    for (int i = 0; i < vertices; i++)
    {
      for (int j = i + 1; j < vertices; j++)
      {
        double p_inv = stats.callsCheck;
        if (matrix[i][j] == adjacent)
        {
          if (edge_stats[i][j])
          {
            p_inv = p_inv / edge_stats[i][j];
          }
          assignmentScore += log2(p_inv);
        }
        else if (matrix[i][j] == not_adjacent)
        {
          if (edge_stats[i][j] != stats.callsCheck)
          {
            p_inv = p_inv / (stats.callsCheck - edge_stats[i][j]);
          }
          assignmentScore += log2(p_inv);
        }
      }
    }

    if (assignmentScore <= assignmentCutoff)
    {
      return false;
    }
  }

  // cutoff reached
  if (!checkMin(frequencyConnectedComponentsSwap != 0))
    return true;
  for (int i = 0; i < vertices; i++)
    for (int j = i + 1; j < vertices; j++)
    {
      if (matrix[i][j] == adjacent)
        printf(" %d", edges[i][j]);
      if (matrix[i][j] == not_adjacent)
        printf(" -%d", edges[i][j]);
    }
  printf("\n");
  vector<lit_t> clause;
  for (int i = 0; i < vertices; i++)
    for (int j = i + 1; j < vertices; j++)
    {
      if (matrix[i][j] == adjacent)
        clause.push_back(-edges[i][j]);
      if (matrix[i][j] == not_adjacent)
        clause.push_back(edges[i][j]);
    }

  // printf("Size %ld\n", clause.size());
  addClause(clause, false);
  return true;
}

adjacency_matrix_t matrixPreviousCall; // the last adjacency matrix where the subgraph is fully defined

bool CommonInterface::checkChromaticNumberSubgraph()
{
  adjacency_matrix_t matrix = getAdjacencyMatrix();
  // printf("Check\n");
  // printAdjacencyMatrix(matrix);

  bool subgraphWasUpdatedandFullyDefined = false;
  if (!subgraphWasUpdatedandFullyDefined)
    for (int i = 0; i < fixedSubgraphSize; i++)
    {
      for (int j = i + 1; j < fixedSubgraphSize; j++)
      {

        if (matrix[i][j] == unknown)
        {
          // printf("Undefined\n");
          return true; // not fully defined so can just return
        }
        // else if ( matrixPreviousCall.size() == 0 || matrix[i][j] != matrixPreviousCall[i][j])
        //  subgraphWasUpdatedandFullyDefined = true;
      }
    }
  matrixPreviousCall = matrix;
  // printf("Check fully defined subgraph\n");
  // printAdjacencyMatrix(matrix);

  auto start = clock();
  // fflush(stdout);

  Coloring c(minChomaticNumberSubgraph - 1);
  coloring_t coloring(fixedSubgraphSize);
  if (c.getColoring(fixedSubgraphSize, matrix, coloring, coloringAlgo, vector<int>()))
  {
    // printf("time for check: %f\n", ((double)clock() - start) / CLOCKS_PER_SEC);
    // printf("Colorable\n");
    auto monoChromaticEdges = c.coloring2monochromaticEdges(coloring);
    vector<lit_t> clause = monochromaticEdges2Clause(monoChromaticEdges);

    if (symBreakClausesFile)
    {
      for (auto edge : monoChromaticEdges)
        fprintf(symBreakClausesFile, "(%d,%d) ", edge.first, edge.second);
      fprintf(symBreakClausesFile, ";coloring\n");
    }
    addClause(clause, false);

    if (addPermutedColorings)
    {
      for (int i = 0; i < (int)coloring.size(); i++)
        for (int j = i + 1; j < (int)coloring.size(); j++)
        {
          if (coloring[i] == coloring[j])
            continue;

          std::swap(coloring[i], coloring[j]); // swap color of vertex i and j
          auto monoChromaticEdges = c.coloring2monochromaticEdges(coloring);
          vector<lit_t> clause = monochromaticEdges2Clause(monoChromaticEdges);

          if (symBreakClausesFile)
          {
            for (auto edge : monoChromaticEdges)
              fprintf(symBreakClausesFile, "(%d,%d) ", edge.first, edge.second);
            fprintf(symBreakClausesFile, ";coloring\n");
          }
          addClause(clause, false);
          std::swap(coloring[i], coloring[j]);
        }
    }
    return false;
  }
  // printf("time for check: %f\n", ((double)clock() - start) / CLOCKS_PER_SEC);
  // printf("Was not colorable.\n");
  return true;
}

bool CommonInterface::propagate()
{
  stats.callsPropagator++;
  auto start = clock();

  bool res;

  if (printPartiallyDefined && rand() % printPartiallyDefined == 0)
  {
    auto matrix = getAdjacencyMatrix();
    printf("p\t");
    printPartiallyDefinedAdjacencyMatrix(matrix);
  }

  if (minChomaticNumberSubgraph) // TODO put in some external function
  {
    res = checkChromaticNumberSubgraph();
  }

  if (thickness2Frequency && rand() % thickness2Frequency == 0)
  {
    res = thicknessTesting();
    // if (!res)
    //   return false;
  }

  if (planarityFrequency && rand() % planarityFrequency == 0)
  {
    res = planarityTesting();
    // if (!res)
    //   return false;
  }

  if (assignmentCutoff && cutoffFunction())
    res = false;
  else if (rand() % frequency != 0)
    res = true;
  else
    res = checkMin(frequencyConnectedComponentsSwap != 0 && rand() % frequencyConnectedComponentsSwap == 0);

  // if (rand() % 10000 == 0)
  //  res = res && extractForbiddenSubgraphs(getAdjacencyMatrix());

  /*if (res && hypercoloring) {
    adjacency_matrix_t matrix = getAdjacencyMatrix();
    bool sufficiently_populated = true;
    for (int e = b_vertices[0]; e < vertices; e++) {
      int e_size = 0;
      for (int v = 0; v < b_vertices[0]; v++) {
        if (matrix[v][e] == adjacent) {
          e_size++;
          if (e_size == 2) {
            break;
          }
        }
      }
      if (e_size < 2) {
        sufficiently_populated = false;
        break;
      }
    }
    if (sufficiently_populated) {
      coloring_t coloring(b_vertices[0]);
      if (getHyperColoring(b_vertices, matrix, coloring)) {
        res = false;
        addClause(getHyperColoringClause(coloring, b_vertices, matrix, edges), false);
        stats.hyperclauses++;
      }
    }
  }*/

  stats.timePropagator += clock() - start;
  return res;
}

bool CommonInterface::extractForbiddenSubgraphs(adjacency_matrix_t matrix)
{
  bool res = true;
  auto subgraphCheckStart = clock();
  // search for all subgraphs simulataniously
  auto subgraphSolver = new CaDiCaL::Solver();
  int varCounter = 1;

  vector<int> selectedGraph;
  vector<vector<int>> permutationVariables; // gives the permutation
  int maxOrder = 0;                         // order of the largest forbidden subgraph
  for (auto subgraph : forbiddenSubgraphs)
  {
    int selectThisGraph = varCounter++;
    selectedGraph.push_back(selectThisGraph);

    int order = 0; // get order of graph
    for (auto e : subgraph)
    {
      order = max(e.first, order);
      order = max(e.second, order);
    }
    order++;

    for (int i = maxOrder; i < order; i++)
    {
      permutationVariables.push_back(vector<int>(vertices));
      iota(permutationVariables[i].begin(), permutationVariables[i].end(), varCounter);
      varCounter += vertices;
    }

    maxOrder = max(order, maxOrder);

    // two different vertices can not be mapped to the same value
    for (int i = 0; i < order; i++)
    {
      for (int j = i + 1; j < order; j++)
      {
        for (int v = 0; v < vertices; v++)
        {
          subgraphSolver->add(-selectThisGraph);
          subgraphSolver->add(-permutationVariables[i][v]);
          subgraphSolver->add(-permutationVariables[j][v]);
          subgraphSolver->add(0);
        }
      }
    }

    // edge must be present if mapped to a certain edge
    for (auto e : subgraph)
    {
      for (int v1 = 0; v1 < vertices; v1++)
      {
        for (int v2 = 0; v2 < vertices; v2++)
        {
          if (v1 == v2 || matrix[v1][v2] == adjacent)
            continue;
          // not permuted to a non-edge
          subgraphSolver->add(-selectThisGraph);
          subgraphSolver->add(-permutationVariables[e.first][v1]);
          subgraphSolver->add(-permutationVariables[e.second][v2]);
          subgraphSolver->add(0);
        }
      }
    }
  }

  // select at least one value to which the vertices mapped if graph; note that more vertices must be mapped although a smaller subgraph then the largest might be selected
  for (auto row : permutationVariables)
  {
    for (auto l : row)
      subgraphSolver->add(l);
    subgraphSolver->add(0);
  }

  // two different vertices can not be mapped to the same value
  for (int i = 0; i < maxOrder; i++)
  {
    for (int j = i + 1; j < maxOrder; j++)
    {
      for (int v = 0; v < vertices; v++)
      {
        subgraphSolver->add(-permutationVariables[i][v]);
        subgraphSolver->add(-permutationVariables[j][v]);
        subgraphSolver->add(0);
      }
    }
  }

  // at least one graph must be selected
  for (auto l : selectedGraph)
    subgraphSolver->add(l);
  subgraphSolver->add(0);

  while (subgraphSolver->solve() == 10) // solving successful
  {
    res = false;
    int subgraph = -1;
    for (int i = 0; i < forbiddenSubgraphs.size(); i++)
      if (subgraphSolver->val(selectedGraph[i]) > 0)
      {
        subgraph = i;
        break;
      }

    // get order of selected graph
    int order = 0; // get order of graph
    for (auto e : forbiddenSubgraphs[subgraph])
    {
      order = max(e.first, order);
      order = max(e.second, order);
    }
    order++;

    // get permutation
    vector<int> permutation;
    for (int i = 0; i < order; i++)
    {
      for (int v = 0; v < vertices; v++)
        if (subgraphSolver->val(permutationVariables[i][v]) > 0)
        {
          permutation.push_back(v);
          break;
        }
    }

    // check just to be sure
    printf("Selected graph:%d with order %d\n", subgraph, order);
    for (auto p : permutation)
      printf("%d ", p);
    printf("\n");

    // forbid graph in other solver as subgraph
    clause_t clause;
    for (auto e : forbiddenSubgraphs[subgraph])
      clause.push_back(-edges[permutation[e.first]][permutation[e.second]]);
    addClause(clause, false);
    printf("Time for subgraph check (SUC): %f\n", ((double)clock() - subgraphCheckStart) / CLOCKS_PER_SEC);

    if (!eliminateAllsubgraphs)
      break;
    // break;
    // forbid the current selection; doesn't consider automorphisms yet
    subgraphSolver->add(-selectedGraph[subgraph]);
    for (int i = 0; i < order; i++)
      subgraphSolver->add(-permutationVariables[i][permutation[i]]);
    subgraphSolver->add(0);
  }
  printf("Time for subgraph check (UNSUC): %f\n", ((double)clock() - subgraphCheckStart) / CLOCKS_PER_SEC);
  delete subgraphSolver;
  return res;
}

// returns false if graph does not satisfy a property
bool CommonInterface::checkFullyDefinedGraph(const adjacency_matrix_t &matrix)
{
  recordGraphStats(matrix);

  if (printFullyDefinedGraphs)
  {
    printf("Check\n");
    fflush(stdout);
    if (hypermode)
    {
      printHypergraph(matrix, b_vertices);
    }
    else
    {
      printAdjacencyMatrix(matrix);
    }
  }

  if (maxPathLength)
  {
    try
    {
      path(matrix);
    }
    catch (const vector<int> path)
    {
      printf("Exclude path with size %ld:", path.size());
      for (auto i : path)
        printf(" %d", i);
      printf("\n");
      if (symBreakClausesFile)
      {
        assert(false); // TODO not implemented yet
      }
      vector<lit_t> clause;
      for (int i = 0; i < (int)path.size() - 1; i++)
        clause.push_back(-edges[path[i]][path[i + 1]]);
      addClause(clause, false);
      return false;
    }
  }

  if (minChromaticNumber)
  {
    Coloring c(minChromaticNumber - 1);
    int nVertices = vertices;

    adjacency_matrix_t intersection_matrix;
    vector<int> clique;
    if (hyperedgeColoring)
    {
      intersection_matrix = getIntersectionMatrix(matrix);
      clique = getCliqueInIntersectionMatrix(matrix);
      // printAdjacencyMatrix(intersection_matrix);
      nVertices = b_vertices[1];
    }
    coloring_t coloring(nVertices);
    if (c.getColoring(nVertices, hyperedgeColoring ? intersection_matrix : matrix, coloring, coloringAlgo, clique))
    {
      // printf("Colorable\n");
      auto monoChromaticEdges = c.coloring2monochromaticEdges(coloring);
      vector<lit_t> clause = monochromaticEdges2Clause(monoChromaticEdges);

      if (symBreakClausesFile)
      {
        for (auto edge : monoChromaticEdges)
          fprintf(symBreakClausesFile, "(%d,%d) ", edge.first, edge.second);
        fprintf(symBreakClausesFile, ";coloring\n");
      }
      addClause(clause, false);

      if (addPermutedColorings)
      {
        for (int i = 0; i < vertices; i++)
          for (int j = i + 1; j < vertices; j++)
          {
            if (coloring[i] == coloring[j])
              continue;

            std::swap(coloring[i], coloring[j]); // swap color of vertex i and j
            auto monoChromaticEdges = c.coloring2monochromaticEdges(coloring);
            vector<lit_t> clause = monochromaticEdges2Clause(monoChromaticEdges);

            if (symBreakClausesFile)
            {
              for (auto edge : monoChromaticEdges)
                fprintf(symBreakClausesFile, "(%d,%d) ", edge.first, edge.second);
              fprintf(symBreakClausesFile, ";coloring\n");
            }
            addClause(clause, false);
            std::swap(coloring[i], coloring[j]);
          }
      }
      return false;
    }
  }
  bool res = true;
  if (forbiddenSubgraphs.size() != 0)
  {
    res = extractForbiddenSubgraphs(matrix);
  }

  if (non010colorable)
  {

    coloring_t coloring(vertices);
    if (get010Coloring(vertices, matrix, coloring))
    {
      /*for (int v = 0; v < vertices; v++) {
        fprintf(stderr, "%d", coloring[v]);
      }
      fprintf(stderr, "\n");*/
      addClause(get010ColoringClause(coloring, vertices, edges, triangles), false);
      if (addPermutedColorings)
      {
        for (int i = 0; i < vertices; i++)
          for (int j = i + 1; j < vertices; j++)
          {
            if (coloring[i] != coloring[j])
            {
              swap(coloring[i], coloring[j]);
              addClause(get010ColoringClause(coloring, vertices, edges, triangles), false);
              swap(coloring[i], coloring[j]); // undo
            }
          }
      }
      return false;
    }
  }

  // get all valid colorings
  if (non010colorableAll)
  {
    auto coloringSolver = new CaDiCaL::Solver();
    vector<int> color(vertices);
    iota(color.begin(), color.end(), 1);

    for (int i = 0; i < vertices; i++)
      for (int j = i + 1; j < vertices; j++)
        if (matrix[i][j] == adjacent)
        {
          coloringSolver->add(color[i]);
          coloringSolver->add(color[j]);
          coloringSolver->add(0);
        }

    for (int i = 0; i < vertices; i++)
      for (int j = i + 1; j < vertices; j++)
        for (int k = j + 1; k < vertices; k++)
          if (matrix[i][j] == adjacent && matrix[i][k] == adjacent && matrix[k][j] == adjacent)
          {
            coloringSolver->add(-color[i]);
            coloringSolver->add(-color[j]);
            coloringSolver->add(-color[k]);
            coloringSolver->add(0);
          }

    int numberOfColorings = 0;
    while (coloringSolver->solve() == 10)
    {
      numberOfColorings++;
      res = false;
      vector<int> resultingColoring(vertices, 0);
      for (int i = 0; i < vertices; i++)
        if (coloringSolver->val(color[i]) < 0)
          resultingColoring[i] = 1;

      addClause(get010ColoringClause(resultingColoring, vertices, edges, triangles), false);

      for (int i = 0; i < vertices; i++)
        if (resultingColoring[i] == 1)
          coloringSolver->add(color[i]);
      coloringSolver->add(0);
    }
    printf("Number of colorings: %d\n", numberOfColorings);
    delete coloringSolver;
  }

  if (hypercoloring)
  {
    coloring_t coloring(b_vertices[0]);
    if (getHyperColoring(b_vertices, matrix, coloring))
    {
      stats.hyperclauses++;
      for (auto &clause : getHyperColoringCircuit(coloring, b_vertices, edges, nextFreeVariable))
      {
        addClause(clause, false);
      }
      return false;
    }
  }
  return res;
}

bool CommonInterface::check()
{
  // printf("Start fully defined\n");
  clock_t start = clock();
  bool res = checkMin(frequencyConnectedComponentsSwap != 0); // check minimality once again before getting fully defined graph
  stats.timePropagator += clock() - start;
  if (!res)
    return false;

  // check properties only tested occasionally
  if (planarityFrequency != 0)
  {
    bool res = planarityTesting();
    if (!res)
      return false;
  }

  if (thickness2Frequency != 0)
  {
    bool res = thicknessTesting();
    if (!res)
      return false;
  }

  // fully defined and minimal graph
  if (printIntermediateStatistic)
  {
    printf("Fully define: %lld\n", stats.callsCheck);
    printStatistics();
  }

  stats.callsCheck++;
  start = clock();
  adjacency_matrix_t matrix = getAdjacencyMatrix();
  // printAdjacencyMatrix(matrix);

  bool r = checkFullyDefinedGraph(matrix);
  stats.timeCheckFullGraphs += clock() - start;
  if (!r)
    return false;

  bool have_solution = !generate_connected || is_connected(matrix, vertices);
  if (have_solution)
  {
    nModels++;
    if (!hideGraphs)
    {

      printf("Solution %d\n", nModels);
      if (hypermode)
      {
        printHypergraph(matrix, b_vertices);
      }
      else
      {
        printAdjacencyMatrix(matrix);
      }
    }
  }

  if (allModels || !have_solution)
  {
    // exclude current graph
    adjacency_matrix_t &m = matrix;
    vector<lit_t> clause;
    if (hyperedgeColoring)
    {
      // forbid not just the specific hypergraph, but also its intersection graph
      adjacency_matrix_t im = getIntersectionMatrix(matrix);
      for (int i = 0; i < b_vertices[1]; i++)
        for (int j = i + 1; j < b_vertices[1]; j++)
          if (im[i][j] == adjacent)
            clause.push_back(-edges_intersection_graph[i][j]);
          else if (im[i][j] == not_adjacent)
            clause.push_back(edges_intersection_graph[i][j]);
          else
            EXIT_UNWANTED_STATE
      addClause(clause, false);
      clause.resize(0);
    }
#ifndef DIRECTED
    for (int i = 0; i < vertices; i++)
      for (int j = i + 1; j < vertices; j++)
        if (m[i][j] == adjacent)
          clause.push_back(-edges[i][j]);
        else if (m[i][j] == not_adjacent)
          clause.push_back(edges[i][j]);
        else
          EXIT_UNWANTED_STATE
    addClause(clause, false);
    return false;
#else
    for (int i = 0; i < vertices; i++)
      for (int j = 0; j < vertices; j++)
        if (i == j)
        {
        }
        else if (m[i][j] == adjacent)
          clause.push_back(-edges[i][j]);
        else if (m[i][j] == not_adjacent)
          clause.push_back(edges[i][j]);
        else
          EXIT_UNWANTED_STATE
    addClause(clause, false);
    return false;
#endif
  }
  return true;
}

void CommonInterface::printStatistics()
{
  if (!printStats)
    return;
  printf("Time in propagator: %f\n", ((double)stats.timePropagator) / CLOCKS_PER_SEC);
  printf("Time in minimality check: %f\n", ((double)stats.timeMinimalityCheck) / CLOCKS_PER_SEC);
  printf("Time in check full graphs: %f\n", ((double)stats.timeCheckFullGraphs) / CLOCKS_PER_SEC);
  printf("Calls of check: %lld\n", stats.callsCheck);
  printf("Calls propagator: %lld\n", stats.callsPropagator);
  printf("Number of hypercolorings: %lld\n", stats.hyperclauses);
  printf("Number of symmetry breaking constraints: %lld\n", stats.nSymBreakClauses);
  if (allModels)
    printf("Number of models: %d\n", nModels);
  // printEdgeStats();
}

void CommonInterface::solve()
{
  // solve
  printf("Start solving\n");
  fflush(stdout);

  // get a solve handle
  int cubeCounter = 0;
  if (!cubeFile.empty())
  {
    ifstream is(cubeFile);
    string line;
    while (getline(is, line))
    {
      cubeCounter++;
      if (rangeCubes.first != 0)
      {
        if (cubeCounter < rangeCubes.first)
          continue;
        if ((cubeCounter > rangeCubes.second))
          break;
      }

      initEdgeMemory(vertices);
      if (non010colorable)
        initTriangleMemory(vertices);
      printf("Solve cube %d\n", cubeCounter);
      vector<lit_t> assumptions; // variable names from the initial encoding

      printf("%s\n", line.c_str());

      istringstream iss(line);

      string lit;
      while (std::getline(iss, lit, ' '))
      {
        if (lit.empty())
          continue;
        assumptions.push_back(stoi(lit));
      }

      clock_t start = clock();

      bool solvedSuccessfully = true;
      if (timeout)
      {
        if (!solve(assumptions, timeout))
          printf("Timeout reached\n");
      }
      else
      {
        solve(assumptions);
      }
      if (solvedSuccessfully)
        printf("Time for cube %f\n", ((double)clock() - start) / CLOCKS_PER_SEC);
      printStatistics();
    }
    printf("All cubes solved\n");
  }
  else
  {
    initEdgeMemory(vertices);
    if (non010colorable)
      initTriangleMemory(vertices);
    if (timeout)
    {
      if (!solve(vector<int>(), timeout))
        printf("Timeout reached\n");
    }
    else
    {
      solve(vector<int>());
    }

    printf("Searched finished\n");
    printStatistics();
  }
}

/*
static inline void greedyConstraint(clingo_propagate_control_t *ctl, int *vertexColoring, propagator_t *data)
{
  nColorings++;
  // printf("START\n");
  int colors = 4;
  int n = data->nVertices;
  int vertexOrdering[n];

  int pos = 0;
  for (int c = 0; c < colors; c++)
    for (int i = 0; i < n; i++)
      if (vertexColoring[i] == c)
        vertexOrdering[pos++] = i;

  assert(pos == n);

  // coloring[i][0] denotes if vertex vertexOrdering[i] has color 0
  clingo_literal_t coloring[n][colors];  // coloring of each vertex
  clingo_literal_t available[n][colors]; // check if color is available, i.e., no smaller vertex in ordering has the color
  clingo_literal_t isColored[n];         // check if vertex is colored
  clingo_literal_t adjacentAndColorC[n][n][colors];

  for (int i = 0; i < n; i++)
  {
    if (!clingo_propagate_control_add_literal(ctl, &isColored[i]))
      exit(EXIT_FAILURE);

    for (int c = 0; c < colors; c++)
    {
      if (!clingo_propagate_control_add_literal(ctl, &coloring[i][c]))
        exit(EXIT_FAILURE);
      if (!clingo_propagate_control_add_literal(ctl, &available[i][c]))
        exit(EXIT_FAILURE);

      for (int j = i + 1; j < n; j++)
      {
        if (!clingo_propagate_control_add_literal(ctl, &adjacentAndColorC[i][j][c]))
          exit(EXIT_FAILURE);

        // printf("%d\n", adjacentAndColorC[i][j][c]);
      }
    }
  }

  // --------------add clauses-----------

  bool res;

  // one color implies isColored
  for (int i = 0; i < n; i++)
  {
    for (int c = 0; c < colors; c++)
    {
      clingo_literal_t clause[] = {-coloring[i][c], isColored[i]};
      addClauseToBuff(clause, 2);
    }
  }

  // at most on color
  for (int i = 0; i < n; i++)
  {
    for (int c1 = 0; c1 < colors; c1++)
    {
      for (int c2 = c1 + 1; c2 < colors; c2++)
      {
        clingo_literal_t clause[] = {-coloring[i][c1], -coloring[i][c2]};
        addClauseToBuff(clause, 2);
      }
    }
  }

  // at least one node is not colored
  clingo_literal_t clause[n];
  for (int i = 0; i < n; i++)
    clause[i] = -isColored[i];
  addClauseToBuff(clause, n);

  // smallest available means that it should get this color
  for (int i = 0; i < n; i++)
  {
    for (int c1 = 0; c1 < colors; c1++)
    {
      clingo_literal_t clause[colors + 2];
      for (int c2 = 0; c2 < c1; c2++) // smaller colors
        clause[c2] = available[i][c2];

      clause[c1] = -available[i][c1];
      clause[c1 + 1] = coloring[i][c1];

      addClauseToBuff(clause, c1 + 2);
    }
  }

  // adjacent and color c
  for (int i = 0; i < n; i++)
  {
    for (int j = i + 1; j < n; j++)
    {
      for (int c = 0; c < colors; c++)
      {

        clingo_literal_t clause1[] = {coloring[i][c], -adjacentAndColorC[i][j][c]}; // not adjacent implies not adjacent and color c
        addClauseToBuff(clause1, 2);

        int v1 = MIN(vertexOrdering[i], vertexOrdering[j]);
        int v2 = MAX(vertexOrdering[i], vertexOrdering[j]);
        clingo_literal_t clause2[2] = {data->edgeLits[v1][v2], -adjacentAndColorC[i][j][c]}; // not adjacent implies not adjacent and color c
        addClauseToBuff(clause2, 2);

        clingo_literal_t clause3[] = {-data->edgeLits[v1][v2], -coloring[i][c], adjacentAndColorC[i][j][c]};
        addClauseToBuff(clause3, 3);
      }
    }
  }

  // if no smaller vertex is adjacent and has color c, than color c is available.
  for (int i = 0; i < n; i++)
  {
    clingo_literal_t clause[i];
    for (int c = 0; c < colors; c++)
    {
      for (int j = 0; j < i; j++)
        clause[j] = adjacentAndColorC[j][i][c];
      clause[i] = available[i][c];
      addClauseToBuff(clause, i + 1);
    }
  }

  // if adjacent and color c than not available
  for (int i = 0; i < n; i++)
  {
    for (int c = 0; c < colors; c++)
    {
      for (int j = 0; j < i; j++)
      {
        // TODO maybe use edge literals and color literals directly.
        clingo_literal_t clause[] = {-adjacentAndColorC[j][i][c], -available[i][c]};
        addClauseToBuff(clause, 2);
      }
    }
  }
  // fflush(stdout);
} */
